/* (C) 1992-2002 by Unicorn Research Corporation, http://www.unicorn.us.com
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * - Neither name of Unicorn Research nor the names of its contributors may be
 * used to endorse or promote products derived from this software without
 * specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE CONTRIBUTORS `AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <locale.h>
#include <langinfo.h>

#define SATURDAY 6               /* 1 Jan 1 was a Saturday                 */
#define FIRST_MISSING_DAY 639799 /* 3 Sep 1752                             */
#define NUMBER_MISSING_DAYS 11   /* 11 day correction during reformation   */
#define is_leap_year(y) \
    ((y) <= 1752 ? !((y) % 4) : \
    (!((y) % 4) && ((y) % 100)) || !((y) % 400))
#define centuries_since_1700(y) \
    ((y) > 1700 ? (y) / 100 - 17 : 0)
#define quad_centuries_since_1700(y) \
    ((y) > 1600 ? ((y) - 1600) / 400 : 0)
#define leap_years_since_year_1(y) \
    ((y) / 4 - centuries_since_1700(y) + quad_centuries_since_1700(y))

short days[] = { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
    thisday, year = 0, europe = 0, maxappts = 10, LineWid = 24,
    day_in_year(short, short, short), day_in_week(short, short, short);

char *months[] = { "", "January ", "February", "March", "April", "May", "June",
         "July", "August " , "September", "October", "November", "December" },
     USdays[] = " Su Mo Tu We Th Fr Sa ",
     *dayline = USdays, *line[10], *buf = NULL, *stryearcpy(char *, short);

void printmonth (short, short);
void fixtab(register short);
void shownow(char *);

void text_init(void)
{
    int i, day;

    strcpy(dayline," ");
    for (i = 0; i < 7; i++) {
        if (europe)
            day = (i + 1) % 7;
        else
            day = i;
    strncat(dayline, nl_langinfo(ABDAY_1+day), 2);
    if (strlen(nl_langinfo(ABDAY_1+day)) == 1)
        strcat(dayline, "  ");
    else
        strcat(dayline, " ");
    }
    for (i = 1; i <= 12; i++) {
        months[i] = nl_langinfo(MON_1+i-1);
    }
}

int main(int argc, char *argv[])
{
    short       bufsiz, i, m, y;
    time_t      t;
    struct tm  *lt;

    setlocale(LC_ALL, "");
    int fwd = (*nl_langinfo(_NL_TIME_FIRST_WEEKDAY) - 1);
    if (fwd == 1) {
        europe = 1;
    } else {
        europe = 0;
    }

    bufsiz = (maxappts)*LineWid;
    if ((buf = (char *)calloc(bufsiz,1)) == NULL) {
        puts("Memory overflow");
        goto freestuff;
    }
    line[0] = buf;
    for (i = 1; i < maxappts; i++)
    line[i] = line[i-1] + LineWid;    /* calendar body    */

    text_init();

    t = time(NULL);
    lt = localtime(&t);
    y = lt->tm_year + 1900;
    m = lt->tm_mon + 1;
    thisday = lt->tm_mday;
    puts("");           /* first display a blank line */
    fixtab (y);
    printmonth(m, y);

    freestuff:
    free(buf);
    return 0;
}

void printmonth(short m, short y)
{
    register short first, last;
    register short index, p = 0;
    register char *ll;
    short l;

    memset(line[0], ' ', (maxappts)*LineWid);
    memcpy(line[0]+((LineWid-2)/2)-((strlen(months[m])+1+4)/2),
           months[m], strlen(months[m]));
    stryearcpy(line[0]+((LineWid-2)/2)-((strlen(months[m])+1
           +4)/2)+strlen(months[m])+1, y);
    line[0][strlen(line[0])] = ' '; /* Remote NULL terminator */
    memcpy(line[1]+p, dayline, LineWid-2);

    /* day of week for first day of month */
    short dow = day_in_week(1, m, y);
    first = 1;
    last = 7 - dow;
    index = 3 * dow;

    for (l = 2; l < 8; ++l) {   /* loop thru month one week per line */
        ll = line[l] + p + 1;
        while (first <= last) {
            if (first >= 10) ll[index] = '0' + first / 10;
            ll[index+1] = '0' + first % 10;
            if (first == thisday) {
                ll[index-1] = '<';
                ll[index+2] = '>';
            }
            index += 3;
            ++first;
        }
        last = (last + 7) > days[m] ? days[m] : last + 7;
        index = 0;
    }
    for (l = 0; l < 8; l++) {
        index = LineWid-1;
        ll = line[l];
        while (index >= 1 && ll[index] == ' ') --index;
        ll[index+1] = '\0';
        shownow(line[l]);
    }
}

void fixtab(short y)
{
    if (is_leap_year(y)) {
        days[2] = 29;
    }
}

short day_in_year(short d, short m, short y)
{
    int i;

    for (i = 1; i < m; i++) {
        if (i == 2)
            d += (is_leap_year(y)) ? 29 : 28;
        else
            d += days[i];
    }
    return d;
}

short day_in_week(short d, short m, short y)
{
    long total_days;
    total_days = (long)(y - 1) * 365 + leap_years_since_year_1(y - 1)
                 + day_in_year(d, m, y);
    return (((total_days - 1 + SATURDAY - europe) - NUMBER_MISSING_DAYS) % 7);
}

void shownow(char *l)
{
    while (*l) {
        fputc(*l++,stdout);
    }
    fputc('\n',stdout);
}

char *stryearcpy(char *dest, short year)
{
    char str[11], *p;

    p = str+10;
    *p = '\0';
    do {
        --p;
        *p = '0' + year % 10;
        year = year / 10;
    } while (year);
    strcpy(dest, p);
    return dest;
}

